"use client";
import {
    createContext,
    useCallback,
    useContext,
    useEffect,
    useMemo,
    useRef,
    useState,
} from "react";
import Button from "react-bootstrap/Button";
import Modal from "react-bootstrap/Modal";
import { TbAntennaBars3 } from "react-icons/tb";
import { Socket } from "socket.io-client";

import type { ID } from "@snip/database/types";
import type {
    ClientToServerEvents,
    ServerToClientEvents,
} from "@snip/socket/types";

import { getSocket } from "./socket";

export interface OnlineUser {
    page_ids: Set<ID> | ID[];
    email: string;
    color: [number, number, number, number];
    id: ID;
    self?: boolean;
}

interface SocketContext {
    socket?: Socket<ServerToClientEvents, ClientToServerEvents>;
    book_id?: ID;
    isConnected: boolean;
    onlineUsers: Map<ID, OnlineUser>;
    pageToOnlineUsers: Map<ID, OnlineUser[]>;
}

const SocketContext = createContext<SocketContext | null>(null);

/** Creates a socket connection to the main
 * namespace of the book. This does not join
 * any pages i.e. rooms. This is intentional
 * as only the worker(s) should join the pages
 * directly.
 * @param book_id - The ID of the book to connect to
 */
export function useSocket(book_id: ID) {
    /** Socket variables
     * normally they are not used but
     * they are here mainly for debugging purposes
     */
    const [socket, setSocket] =
        useState<Socket<ServerToClientEvents, ClientToServerEvents>>();
    const [isConnected, setIsConnected] = useState(false);

    useEffect(() => {
        // Create the socket
        getSocket(book_id).then((newSocket) => {
            setSocket((socket) => {
                if (socket) {
                    socket.close();
                }
                return newSocket;
            });
        });
        return () => {
            // Close the socket
            getSocket(book_id)?.then((s) => s.close());
        };
    }, [book_id]);

    // Register connection and disconnection events
    useEffect(() => {
        if (!socket) return;

        function onConnect() {
            console.log("[Socket] Connected to book", book_id);
            setIsConnected(true);
        }

        function onDisconnect() {
            console.log("[Socket] Disconnected from book", book_id);
            setIsConnected(false);
        }

        function reConnect() {
            if (socket?.connected !== true) {
                socket?.connect();
            }
        }

        socket.on("connect", onConnect);
        socket.on("disconnect", onDisconnect);
        socket.connect();
        window.addEventListener("online", reConnect);
        window.addEventListener("visibilitychange", reConnect);

        return () => {
            socket.off("connect", onConnect);
            socket.off("disconnect", onDisconnect);
            window.removeEventListener("online", reConnect);
            window.removeEventListener("visibilitychange", reConnect);
            socket.disconnect();
        };
    }, [book_id, socket]);

    return { socket, isConnected };
}

export function useOnlineUsers(
    isConnected: boolean,
    socket?: Socket<ServerToClientEvents, ClientToServerEvents>,
) {
    /** Other users currently online and viewing pages
     * of the book
     */
    const [onlineUsers, setOnlineUsers] = useState<Map<ID, OnlineUser>>(
        new Map(),
    );

    //Register online users after connection
    useEffect(() => {
        if (!isConnected || !socket) return;

        socket.emit("users:getOnline", (users) => {
            if (users.ok === false) {
                console.error(users.error);
                return;
            }

            console.log("[Socket] Online users", users);
            setOnlineUsers(
                users.data.reduce(
                    (map, user) => map.set(user.id, user),
                    new Map(),
                ),
            );
        });

        function onUserDisconnect(user_id: ID) {
            setOnlineUsers((users) => {
                users.delete(user_id);
                return new Map(users);
            });
        }

        function onUserUpdate(user: OnlineUser) {
            console.log("[Socket] User update", user);
            setOnlineUsers((users) => {
                //only update if the page_ids are different
                if (!users.has(user.id)) {
                    return new Map(users).set(user.id, user);
                }
                if (
                    !user.page_ids ||
                    users.get(user.id)?.page_ids !== user.page_ids
                ) {
                    return new Map(users).set(user.id, user);
                }
                return users;
            });
        }

        socket.on("users:disconnect", onUserDisconnect);
        socket.on("users:update", onUserUpdate);
        socket.on("users:connect", onUserUpdate);

        return () => {
            socket.off("users:disconnect", onUserDisconnect);
            socket.off("users:update", onUserUpdate);
            socket.off("users:connect", onUserUpdate);
        };
    }, [socket, isConnected]);

    return { onlineUsers };
}

export function SocketContextProvider({
    children,
    book_id,
}: {
    children: React.ReactNode;
    book_id: ID;
}) {
    const { socket, isConnected } = useSocket(book_id);
    const { onlineUsers } = useOnlineUsers(isConnected, socket);

    // Map page ids to online users
    const pageToOnlineUsers = new Map<ID, OnlineUser[]>();
    onlineUsers?.forEach((user) => {
        user?.page_ids?.forEach((page_id) => {
            let new_users = pageToOnlineUsers.get(page_id) || [];
            new_users.push(user);

            // filter out duplicates
            new_users = new_users.filter(
                (user, index, self) =>
                    index === self.findIndex((u) => u.id === user.id),
            );

            pageToOnlineUsers.set(page_id, new_users);
        });
    });

    return (
        <SocketContext.Provider
            value={{
                socket,
                isConnected,
                book_id,
                onlineUsers,
                pageToOnlineUsers,
            }}
        >
            <LatencyProblemsModal />
            {children}
        </SocketContext.Provider>
    );
}

function LatencyProblemsModal() {
    const { socket } = useSocketContext();

    const [history, setHistory] = useState<number[]>([]);

    const latencyCheck = useCallback(() => {
        const start = Date.now();
        socket?.emit("ping", () => {
            const total = Date.now() - start;
            setHistory((history) => {
                if (history.length > 1000) {
                    history.shift();
                }
                return [...history, total];
            });
        });
    }, [socket]);

    useEffect(() => {
        const interval = setInterval(latencyCheck, 4000); //every 4 seconds by default
        return () => clearInterval(interval);
    }, [latencyCheck]);

    // Show the mode if the history is longer than 5 and the mean is greater than 500ms
    const [_show, setShow] = useState<boolean | null>(null);
    const mean = useMemo(() => {
        if (history.length < 5) return 0;
        const sortedHistory = [...history].sort((a, b) => a - b);
        const middleIndex = Math.floor(sortedHistory.length / 2);
        const median =
            sortedHistory.length % 2 === 0
                ? (sortedHistory[middleIndex - 1]! +
                      sortedHistory[middleIndex]!) /
                  2
                : sortedHistory[middleIndex];
        return median || 0;
    }, [history]);

    function toggle() {
        setShow(!_show);
    }

    let show;
    if (_show === null) {
        show = mean > 500;
    } else {
        show = _show;
    }

    return (
        <>
            <Modal
                show={show}
                onHide={toggle}
                animation={false}
                backdrop="static"
            >
                <Modal.Header closeButton>
                    <Modal.Title>Unstable Connection Detected !</Modal.Title>
                </Modal.Header>
                <Modal.Body className="d-flex flex-column">
                    <p>
                        We have detected an unstable internet connection. Please
                        check your connection. We can not guarantee that your
                        changes will be reflected in real-time!
                    </p>
                    <div>
                        Current Latency: <code>{JSON.stringify(mean)}ms</code>
                    </div>
                    <LatencyGraph history={history} />
                </Modal.Body>
                <Modal.Footer>
                    <Button
                        variant="secondary"
                        onClick={() => {
                            // Check the latency again for 10 seconds each
                            // second
                            const interval = setInterval(latencyCheck, 50);
                            setTimeout(() => clearInterval(interval), 10000);
                        }}
                    >
                        Run test
                    </Button>
                    <Button variant="primary" onClick={toggle}>
                        Ignore
                    </Button>
                </Modal.Footer>
            </Modal>
            {mean > 250 && (
                <TbAntennaBars3
                    style={{
                        position: "fixed",
                        bottom: "0",
                        right: "0",
                        zIndex: 9999,
                        margin: "0.2rem",
                        borderRadius: "50%",
                        padding: "0rem",
                        aspectRatio: "1/1",
                        display: "flex",
                        alignItems: "center",
                        justifyContent: "center",
                        backgroundColor: "white",
                        border: "2px solid var(--bs-danger)",
                        color: "var(--bs-danger)",
                    }}
                    size={28}
                    onClick={toggle}
                />
            )}
        </>
    );
}

export function useSocketContext() {
    const context = useContext(SocketContext);
    if (!context) throw new Error("Cant use SocketContext outside provider!");
    return context;
}

interface LatencyGraphProps {
    history: number[];
}

const LatencyGraph: React.FC<LatencyGraphProps> = ({ history }) => {
    const svgRef = useRef<SVGSVGElement>(null);

    useEffect(() => {
        const svg = svgRef.current;
        if (!svg) return;

        const width = svg.clientWidth;
        const height = svg.clientHeight;
        const maxHistoryValue = Math.max(...history);

        const xScale = (index: number) =>
            (index / (history.length - 1)) * width;
        const yScale = (value: number) =>
            height - (value / maxHistoryValue) * height;

        const line = document.createElementNS(
            "http://www.w3.org/2000/svg",
            "path",
        );
        line.setAttribute("fill", "none");
        line.setAttribute("stroke", "var(--bs-primary)");
        line.setAttribute("stroke-width", "2");

        // Create a gradient element
        const defs = document.createElementNS(
            "http://www.w3.org/2000/svg",
            "defs",
        );

        const gradient = document.createElementNS(
            "http://www.w3.org/2000/svg",
            "linearGradient",
        );
        gradient.setAttribute("id", "lineGradient");
        gradient.setAttribute("x1", "0%");
        gradient.setAttribute("x2", "0%");
        gradient.setAttribute("y1", "0%");
        gradient.setAttribute("y2", "100%");

        // Define gradient stops
        const lowValue = 30;
        const highValue = 500;

        const lowStop = document.createElementNS(
            "http://www.w3.org/2000/svg",
            "stop",
        );
        lowStop.setAttribute(
            "offset",
            `${(1 - lowValue / maxHistoryValue) * 100}%`,
        );
        lowStop.setAttribute("stop-color", "red");

        const highStop = document.createElementNS(
            "http://www.w3.org/2000/svg",
            "stop",
        );
        highStop.setAttribute(
            "offset",
            `${(1 - highValue / maxHistoryValue) * 100}%`,
        );
        highStop.setAttribute("stop-color", "green");

        gradient.appendChild(lowStop);
        gradient.appendChild(highStop);

        defs.appendChild(gradient);

        const pathData = history
            .map((value, index) => `${xScale(index)},${yScale(value)}`)
            .join(" L ");

        // Ensure the path starts with 'M' and follows with 'L' for the subsequent points
        line.setAttribute("d", `M ${pathData}`);

        // Apply the gradient to the line
        line.setAttribute("stroke", "url(#lineGradient)");

        // Clear any previous lines before appending the new one
        while (svg.firstChild) {
            svg.removeChild(svg.firstChild);
        }
        svg.appendChild(defs);
        svg.appendChild(line);
    }, [history]);

    return <svg ref={svgRef} style={{ width: "100%", height: "100%" }}></svg>;
};
