"use client";
import Image from "next/image";
import { useEffect, useRef } from "react";

import { isInViewport } from "@snip/common";
import { PageData } from "@snip/database/types";

import { useSidebarContext } from "../context";

import { usePagesContext } from "components/context/socket/usePagesContext";
import {
    OnlineUser,
    useSocketContext,
} from "components/context/useSocketContext";
import { useEditorContext } from "components/editor/context";

import styles from "./preview.module.scss";

/** The page previews component shows
 * a preview of all pages in the book.
 * Each page is an image as rendered
 * by the render service.
 *
 * On an update the
 *
 * It allows to pass a filter to only show
 * a subset of the pages.
 */
export function PreviewPages() {
    const { pagesArray: pages } = usePagesContext();
    const { filterPages } = useSidebarContext();
    const { activePageIds, setActivePageIds } = useEditorContext();
    const { pageToOnlineUsers } = useSocketContext();

    function onPreviewClick(
        page: PageData,
        e: React.MouseEvent<HTMLDivElement, MouseEvent>,
    ) {
        e.preventDefault();

        if (e.ctrlKey || e.shiftKey) {
            // add page to active pages up to max 2
            setActivePageIds((prev) => {
                if (!prev.includes(page.id) && prev.length < 2) {
                    prev.push(page.id);
                } else if (prev.includes(page.id)) {
                    prev = prev.filter((p) => p !== page.id);
                } else if (!prev.includes(page.id)) {
                    //pop push new
                    prev = prev.filter((p) => p !== page.id);
                    prev.pop();
                    prev.unshift(page.id);
                }
                return [...prev];
            });
        } else {
            // set page as active
            setActivePageIds((prev) => {
                if (prev.includes(page.id)) {
                    return prev;
                }
                prev.pop();
                prev.unshift(page.id);
                return [...prev];
            });
        }
    }

    return (
        <div className={styles.wrapper}>
            {pages.map((page, i) => {
                return (
                    <PreviewSinglePage
                        page={page}
                        key={i}
                        hidden={
                            filterPages ? !filterPages.includes(page.id) : false
                        }
                        onClick={(page, e) => {
                            onPreviewClick(page, e);
                        }}
                        active={activePageIds?.includes(page.id)}
                        onlineUsers={pageToOnlineUsers.get(page.id) || []}
                    />
                );
            })}
        </div>
    );
}

/**
 * Renders a single page preview.
 *
 * @param page - The page data.
 * @param active - Indicates if the page is active.
 * @param hidden - Indicates if the page is hidden.
 * @param onClick - The function to call when the page is clicked.
 */
export function PreviewSinglePage({
    page,
    active = false,
    hidden = false,
    onClick,
    onlineUsers = [],
}: {
    page: PageData;
    active?: boolean;
    hidden?: boolean;
    onClick?: (
        page: PageData,
        e: React.MouseEvent<HTMLDivElement, MouseEvent>,
    ) => void;
    onlineUsers: OnlineUser[];
}) {
    // Scroll to page preview when active
    const ref = useRef<HTMLImageElement>(null);

    useEffect(() => {
        if (active && ref.current && !isInViewport(ref.current)) {
            ref.current.scrollIntoView({
                behavior: "smooth",
                block: "nearest",
                inline: "nearest",
            });
        }
    }, [active]);

    return (
        <div
            className={styles.page}
            data-active={active}
            data-hidden={hidden}
            onClick={(e) => onClick?.(page, e)}
            ref={ref}
        >
            <ImagePagePreview page={page} />
            <label>{page.page_number! + 1}</label>
            {onlineUsers.length > 0 && (
                <div className={styles.onlineUsers}>
                    {onlineUsers.map((user, i) => {
                        if (i == 6) {
                            return (
                                <label
                                    key={i}
                                    title={
                                        "" + (onlineUsers.length - 6) + " more"
                                    }
                                >
                                    ..
                                </label>
                            );
                        }
                        if (i > 6) return;

                        return (
                            <label
                                key={i}
                                title={user?.email || ""}
                                style={{
                                    color: `rgba(${user.color.join(",")})`,
                                    borderColor: `rgba(${user.color.join(",")})`,
                                }}
                            >
                                {user.email[0]?.toLocaleUpperCase()}
                            </label>
                        );
                    })}
                </div>
            )}
        </div>
    );
}

function ImagePagePreview({ page }: { page: PageData }) {
    return (
        <Image
            src={page.id.toString()}
            alt="Page"
            width={150}
            height={150 * 1.41}
            loader={({ src, width }) => {
                // 200 hardcoded for now
                return `/render/${src}.jpeg?w=${width}&date=${page.last_updated?.valueOf()}`;
            }}
        />
    );
}
